"""Configuration tab for the Gradio app."""

import os
import gradio as gr
import yaml
from typing import Tuple
from loguru import logger

from toxicity_detector.config import PipelineConfig
from toxicity_detector.app.app_config_loader import _app_config, _config_manager

# TODO: Use decorators consistently for event handling.


def create_config_tab(
    tw_approved: gr.State,
    tw_approved_value: bool,
    pipeline_config_state: gr.State,
) -> gr.Tab:
    """Create the configuration tab.

    Args:
        tw_approved: State for trigger warning approval
        tw_approved_value: Boolean value indicating if trigger warning was approved
        pipeline_config_state: State for pipeline configuration

    Returns:
        The configuration tab component
    """
    with gr.Tab(
        label="Konfiguration",
        id="config_tab",
        visible=tw_approved_value and _app_config().developer_mode,
    ) as config_tab:
        with gr.Row():
            with gr.Column(scale=4, min_width=300):
                yaml_config_input = gr.Code(
                    _config_manager().get_pipeline_config_as_string(),
                    language="yaml",
                    interactive=True,
                )
            with gr.Column(scale=1, min_width=50):
                dropdown_config = gr.Dropdown(
                    choices=_config_manager().list_config_files(),
                    value=os.path.basename(
                        _config_manager().default_pipeline_config_file
                    ),
                    allow_custom_value=False,
                    label="Konfigurationsdatei",
                    info="Wähle die zu ladende Konfigurationsdatei aus!",
                    interactive=True,
                )
                reload_config_btn = gr.Button("Eingegebene Konfiguration laden")
                with gr.Group():
                    gr.Markdown("  Speichern der aktuellen Konfiguration.")
                    new_config_name_tb = gr.Textbox(
                        label="Dateiname (mit Dateiendung .yaml)",
                    )
                    save_config_btn = gr.Button("Speichern")

        # EVENT LISTENER/LOGIC FOR CONFIG TAB
        @reload_config_btn.click(  # RELOAD CONFIG BUTTON
            inputs=yaml_config_input, outputs=pipeline_config_state
        )
        def parse_yaml_str(yaml_str: str) -> PipelineConfig:
            logger.debug("Loading pipeline config from YAML string...")
            try:
                config_dict = yaml.safe_load(yaml_str)
                return PipelineConfig(**config_dict)
            except yaml.YAMLError as e:
                raise gr.Error(f"Error parsing YAML: {e}")

        @dropdown_config.input(
            inputs=dropdown_config,
            outputs=[yaml_config_input, pipeline_config_state],
            show_progress="minimal",
        )
        def load_selected_config(  # type: ignore
            config_file_name: str,
        ) -> Tuple[str, PipelineConfig]:
            logger.info(f"Loading selected config: {config_file_name}")
            pipeline_config = _config_manager().load_pipeline_config(config_file_name)
            pipeline_config_str = _config_manager().get_pipeline_config_as_string(
                config_file_name
            )
            return pipeline_config_str, pipeline_config

        # SAVE-PIPELINE-CONFIG BUTTON
        @save_config_btn.click(
            inputs=[new_config_name_tb, yaml_config_input],
            outputs=[dropdown_config, pipeline_config_state],
        )
        def save_config(new_config_name: str, config_str: str):
            if not new_config_name or new_config_name.isspace():
                raise gr.Error(
                    "Der Name der neuen Konfiguration " "darf nicht leer sein."
                )
            try:
                logger.debug("Loading pipeline config from YAML string...")
                config = PipelineConfig(**yaml.safe_load(config_str))
            except yaml.YAMLError as e:
                raise gr.Error(f"Error parsing YAML: {e}")
            # save str from yaml_config_input as file
            new_config_file_name = f"{new_config_name}.yaml"
            if _config_manager().file_exists(new_config_file_name):
                raise gr.Error(
                    f"Eine Konfigurationsdatei mit dem Name "
                    f"{new_config_file_name} existiert schon."
                )
            _config_manager().save_string(new_config_name, config_str)
            # Update the dropdown with the new config file
            return (
                gr.Dropdown(
                    choices=_config_manager().list_config_files(),
                    value=new_config_file_name,
                    interactive=True,
                ),
                config,
            )

    return config_tab



from pydantic import (
    BaseModel
)
import enum
from typing import Dict, Optional, List


class ToxicityType(enum.Enum):
    GRUP = "hatespeech"
    PERS = "personalized_toxicity"


class ToxicityAnswer(enum.Enum):
    """Enum for toxicity detection results."""
    TRUE = "true"  # Text contains toxic content
    FALSE = "false"  # Text does not contain toxic content
    UNCLEAR = "unclear"  # Insufficient information to determine


class Task(BaseModel):
    name: str
    llm_description: str
    positive_examples: Optional[List[str]] = None
    negative_examples: Optional[List[str]] = None


class Toxicity(BaseModel):
    title: str
    user_description: str
    llm_description: str
    tasks: Dict[str, Dict[str, Task]]



"""Command-line interface for the Toxicity Detector."""

import argparse
import sys
from pathlib import Path
from loguru import logger

from toxicity_detector import detect_toxicity, PipelineConfig, AppConfig
from toxicity_detector.datamodels import ToxicityType


def detect_command(args):
    """Run toxicity detection on input text."""
    # Load pipeline configuration
    try:
        pipeline_config = PipelineConfig.from_file(args.pipeline_config)
    except Exception as e:
        logger.error(
            f"Failed to load pipeline config from '{args.pipeline_config}': {e}"
        )
        sys.exit(1)

    # Validate toxicity type
    if args.toxicity_type not in ToxicityType._value2member_map_:
        logger.error(
            f"Invalid toxicity type '{args.toxicity_type}'. "
            f"Valid options: {list(ToxicityType._value2member_map_.keys())}"
        )
        sys.exit(1)

    # Run toxicity detection
    try:
        result = detect_toxicity(
            input_text=args.text,
            user_input_source=args.source,
            toxicity_type=args.toxicity_type,
            context_info=args.context,
            pipeline_config=pipeline_config,
            serialize_result=args.save,
        )

        # Display results
        print("\n" + "=" * 80)
        print("TOXICITY DETECTION RESULT")
        print("=" * 80)
        print(f"\nInput text: {args.text}")
        print(f"Toxicity type: {args.toxicity_type}")
        print(f"\nContains toxicity: {result.answer.get('contains_toxicity', 'N/A')}")

        if args.verbose:
            print(f"\nFull analysis: {result.answer.get('analysis_result', 'N/A')}")

        if args.save:
            print(f"\nResult saved to: {pipeline_config.result_data_path}")
        print("=" * 80 + "\n")

    except Exception as e:
        logger.error(f"Error during toxicity detection: {e}")
        if args.verbose:
            logger.exception("Full traceback:")
        sys.exit(1)


def app_command(args):
    """Start the Gradio app."""
    # Import here to avoid loading Gradio unless needed
    from toxicity_detector.app import launch_app

    config_path = args.app_config or args.pipeline_config
    config_type = "app" if args.app_config else "pipeline"
    logger.info(f"Starting app with {config_type} config: {config_path}")

    try:
        launch_app(
            config_path=config_path,
            config_type=config_type,
            server_name=args.server_name,
            server_port=args.server_port,
            share=args.share,
        )
    except Exception as e:
        logger.error(f"Failed to start app: {e}")
        if args.verbose:
            logger.exception("Full traceback:")
        sys.exit(1)


def main():
    """Main CLI entry point."""
    parser = argparse.ArgumentParser(
        prog="toxicity-detector",
        description="LLM-based pipeline to detect toxic speech",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  # Run toxicity detection on text
  toxicity-detector detect --text "Your text here" --pipeline-config config/pipeline_config.yaml

  # Start the Gradio app with app config
  toxicity-detector app --app-config config/app_config.yaml

  # Start the Gradio app with pipeline config (uses default app settings)
  toxicity-detector app --pipeline-config config/pipeline_config.yaml --share
        """,
    )

    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Detect subcommand
    detect_parser = subparsers.add_parser(
        "detect",
        help="Run toxicity detection on input text",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    detect_parser.add_argument(
        "--text",
        "-t",
        required=True,
        help="Text to analyze for toxicity",
    )
    detect_parser.add_argument(
        "--pipeline-config",
        "-p",
        required=True,
        help="Path to pipeline configuration YAML file (required)",
    )
    detect_parser.add_argument(
        "--toxicity-type",
        "-T",
        default="hatespeech",
        help="Type of toxicity to detect 'personlized_toxicity' or 'hatespeech' (default: 'hatespeech')",
    )
    detect_parser.add_argument(
        "--source",
        "-s",
        default=None,
        help="Source identifier for the input (e.g., 'chat', 'comment')",
    )
    detect_parser.add_argument(
        "--context",
        "-c",
        default=None,
        help="Additional context about the conversation or situation",
    )
    detect_parser.add_argument(
        "--save",
        action="store_true",
        default=True,
        help="Save the result to disk as YAML",
    )
    detect_parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Show detailed output",
    )
    detect_parser.set_defaults(func=detect_command)

    # App subcommand
    app_parser = subparsers.add_parser(
        "app",
        help="Start the Gradio web interface",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    config_group = app_parser.add_mutually_exclusive_group(required=True)
    config_group.add_argument(
        "--app-config",
        "-a",
        help="Path to app configuration YAML file",
    )
    config_group.add_argument(
        "--pipeline-config",
        "-p",
        help="Path to pipeline configuration YAML file (uses default app settings)",
    )

    app_parser.add_argument(
        "--server-name",
        default="127.0.0.1",
        help="Server name (default: 127.0.0.1)",
    )
    app_parser.add_argument(
        "--server-port",
        type=int,
        default=7860,
        help="Server port (default: 7860)",
    )
    app_parser.add_argument(
        "--share",
        action="store_true",
        help="Create a public shareable link",
    )
    app_parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Show detailed output",
    )
    app_parser.set_defaults(func=app_command)

    # Parse arguments
    args = parser.parse_args()

    # Show help if no command specified
    if not args.command:
        parser.print_help()
        sys.exit(1)

    # Execute the appropriate command
    args.func(args)


if __name__ == "__main__":
    main()



import pydantic
import uuid
from typing import Any, Dict
from datetime import datetime, timezone

from toxicity_detector.config import PipelineConfig


class ToxicityDetectorResult(pydantic.BaseModel):
    user_input: str | None = None
    user_input_source: str | None = None
    context_information: str | None = None
    toxicity_type: str | None = None
    answer: Dict[str, Any] = {}
    pipeline_config: PipelineConfig | None = None
    feedback: Dict[str, Any] = {}
    request_id: str = pydantic.Field(
        default_factory=lambda: str(uuid.uuid4())
    )
    request_time: str = pydantic.Field(
        default_factory=lambda: datetime.now(
            timezone.utc
        ).strftime("%Y-%m-%d %H:%M:%S UTC"))



"""Toxicity Detector - An LLM-based pipeline to detect toxic speech."""

# Import from chains module
from .chains import (
    BaseChainBuilder,
    IdentifyToxicContentZeroShotChain,
    MonoModelDetectToxicityChain,
    IdentifyToxicContentChatChain,
)

# Import from backend module
from .backend import (
    detect_toxicity,
    ZeroShotClassifier,
    update_feedback,
    save_result,
    get_openai_chat_model,
)

from .result import (
    ToxicityDetectorResult,
)

from .config import (
    AppConfig,
    PipelineConfig,
    SubdirConstruction,
)

from .managers import (
    ConfigManager,
)

from .datamodels import (
    ToxicityType,
    ToxicityAnswer,
    Toxicity,
    Task
)

__all__ = [
    # Chain classes
    "BaseChainBuilder",
    "IdentifyToxicContentZeroShotChain",
    "MonoModelDetectToxicityChain",
    "IdentifyToxicContentChatChain",
    # Backend classes
    "ZeroShotClassifier",
    # Backend functions
    "detect_toxicity",
    "update_feedback",
    "save_result",
    "get_openai_chat_model",
    # Config, output and other basic classes
    "ToxicityType",
    "ToxicityAnswer",
    "Toxicity",
    "Task",
    "ToxicityDetectorResult",
    "PipelineConfig",
    "AppConfig",
    "SubdirConstruction",
    "ConfigManager",
]



from glob import glob
import os
from typing import Any, Dict, List, Optional

from langchain_core.callbacks.manager import CallbackManagerForLLMRun
from langchain_core.language_models.llms import LLM
from langchain_openai import ChatOpenAI
from pydantic import SecretStr
from huggingface_hub import InferenceClient
import yaml
from huggingface_hub import HfFileSystem
from datetime import datetime
import pandas as pd
from loguru import logger

from toxicity_detector import MonoModelDetectToxicityChain
from toxicity_detector.result import ToxicityDetectorResult
from toxicity_detector.config import AppConfig, SubdirConstruction, PipelineConfig
from toxicity_detector.datamodels import ToxicityAnswer


def detect_toxicity(
    input_text: str,
    user_input_source: str | None,
    toxicity_type: str,  # TODO: use enum
    context_info: str | None,
    pipeline_config: PipelineConfig,
    serialize_result: bool = True,
):
    result = ToxicityDetectorResult(
        user_input=input_text,
        user_input_source=user_input_source,
        toxicity_type=toxicity_type,  # TODO: use enum
        context_information=context_info,
        pipeline_config=pipeline_config,
    )

    if not input_text or input_text == "":
        raise ValueError("Input text must not be empty.")

    # create new uuid for the detection request
    # (used for UI logic to attach user feedback and for data serialization)

    logger.info(f"Starting new detection request (uuid: {result.request_id}).")
    context_info = None if not context_info or context_info.isspace() else context_info
    model = pipeline_config.used_chat_model

    logger.info(f"Chosen toxicity type: {toxicity_type}")
    logger.info(f"Used model: {pipeline_config.models[model]['name']}")
    logger.info(f"Kontextinfo: {context_info}")
    # Chat model
    if pipeline_config.models[model]["llm_chain"] == "chat-chain":
        # getting api key
        if "api_key" in pipeline_config.models[model].keys():
            api_key = SecretStr(pipeline_config.models[model]["api_key"])
        elif "api_key_name" in pipeline_config.models[model].keys():
            api_key_name = pipeline_config.models[model]["api_key_name"]
            logger.info(f"Used api key name: {api_key_name}")
            # check whether the api key is set as env variable
            if os.environ.get(api_key_name) is None:
                raise ValueError(
                    f"The api key name {api_key_name} is not set as " f"env variable."
                )
            api_key = SecretStr(os.environ.get(api_key_name, "no-api-key"))
        else:
            raise ValueError(
                "You should specify in the config yaml either an api key "
                "or an api-key name (if it is to be found as env variable)."
            )
        if api_key is None:
            raise ValueError(
                "You should specify an api key (recommended: as env variable)."
            )
        # model params
        model_kwargs = {}
        if "model_kwargs" in pipeline_config.models[model].keys():
            model_kwargs = pipeline_config.models[model]["model_kwargs"]

        logger.info(f"Model kwargs: {model_kwargs}")
        # building chain
        toxicitiy_detection_chain = MonoModelDetectToxicityChain.build(
            llms_dict={
                "chat_model": get_openai_chat_model(
                    api_key=api_key,
                    model=pipeline_config.models[model]["model"],
                    base_url=pipeline_config.models[model]["base_url"],
                )
            },
            preprocessing=pipeline_config.get_prompt_messages("preprocessing"),
            indicator_classification=pipeline_config.get_prompt_messages(
                "indicator_classification"
            ),
            indicator_aggregation=pipeline_config.get_prompt_messages(
                "indicator_aggregation"
            ),
            formatting_prompt_msgs=pipeline_config.get_prompt_messages(
                "formatting_prompt_msgs"
            ),
            indicators_dict={
                key: pipeline_config.toxicities[toxicity_type]
                .tasks["indicator_analysis"][key]
                .model_dump()
                for key in pipeline_config.toxicities[toxicity_type]
                .tasks["indicator_analysis"]
                .keys()
            },
            **model_kwargs,
        )
    else:
        err_msg = (
            f"llm_chain "
            f"{pipeline_config.models[model]['llm_chain']} not "
            f"implemented."
        )
        logger.error(err_msg)
        raise ValueError(err_msg)
    # TODO: Offering possibility to use zero shot classifiers
    # (for, e.g., indicator analysis)
    # using zero-shot classifier
    # elif pipeline_config.models[model]['llm_chain'] == \
    #         "zero-shot-chain":
    #     # Identifier of the model (as used in the config)
    #     # that is being used to provide an explanation for the
    #     # zero-shot categorisation.
    #     explaining_model_name = model_config_dict['toxicities'][
    #         toxicity_type]['zero-shot-categorization'][
    #         'explaining_model']
    #     # The "chain" for justifying/explaining the categorization
    #     log_msg(model_config_dict['toxicities'][toxicity_type][
    #         'zero-shot-categorization']['labels'].values())
    #     toxicity_detection_chain = \
    #         chains.IdentifyToxicContentZeroShotChain.build({
    #         'zero_shot_model': backend.ZeroShotClassifier(
    #             model=pipeline_config.models[model]['repo_id'],
    #             labels=list(model_config_dict['toxicities'][
    #                 toxicity_type]['zero-shot-categorization'][
    #                 'labels'].values()),
    #             multi_label=model_config_dict['toxicities'][toxicity_type]['zero-shot-categorization']['multi_label'],
    #             hypothesis_template=model_config_dict['toxicities'][toxicity_type]['zero-shot-categorization']['hypothesis_template'],
    #             api_token=SecretStr(os.environ.get("HF_TOKEN_KIDEKU_INFERENCE"))
    #         ),
    #         'chat_model': backend.get_chat_model(
    #             token=os.environ.get("HF_TOKEN_KIDEKU_INFERENCE"),
    #             repo_id=pipeline_config.models[explaining_model_name]['repo_id']
    #         )
    #     })

    indicators_dict = {
        key: pipeline_config.toxicities[toxicity_type]
        .tasks["indicator_analysis"][key]
        .model_dump()
        for key in pipeline_config.toxicities[toxicity_type]
        .tasks["indicator_analysis"]
        .keys()
    }

    answer = toxicitiy_detection_chain.invoke(
        {
            "system_prompt": pipeline_config.system_prompt,
            "toxicity_explication": pipeline_config.toxicities[
                toxicity_type
            ].llm_description,
            "user_input": input_text,
            "user_input_source": user_input_source,
            "general_questions": pipeline_config.toxicities[toxicity_type]
            .tasks["prepatory_analysis"]["general_questions"]
            .model_dump(),
            "context_information": context_info,
            "indicators_dict": indicators_dict,
        }
    )

    result.answer = answer
    # adding prompts for logging
    prompts = MonoModelDetectToxicityChain.prompts(
        preprocessing=pipeline_config.get_prompt_messages("preprocessing"),
        indicator_classification=pipeline_config.get_prompt_messages(
            "indicator_classification"
        ),
        indicator_aggregation=pipeline_config.get_prompt_messages(
            "indicator_aggregation"
        ),
        **result.answer,
    )
    result.answer["prompts"] = prompts
    # saving the result
    # TODO: as async
    if serialize_result:
        save_result(
            result=result,
            pipeline_config=pipeline_config,
        )

    return result


class ZeroShotClassifier(LLM):
    """LLM wrapper for zero shot classification via HF API."""

    model: str
    """Model ID or deployed Endpoint URL for inference."""

    labels: List[str]
    """List of label verbalizations for input text."""

    multi_label: bool = False
    """If True, labels evaluated independently. If False, mutually
    exclusive (sum to 1)."""

    hypothesis_template: str | None = None
    """
    A template sentence string with curly brackets to which the label strings are added. The label strings are added at the position of the curly brackets ”{}“. Zero-shot classifiers are based on NLI models, which evaluate if a hypothesis is entailed in another text or not. For example, with hypothesis_template=“This text is about {}.” and labels=[“economics”, “politics”], the system internally creates the two hypotheses “This text is about economics.” and “This text is about politics.”. The model then evaluates for both hypotheses if they are entailed in the provided text or not. # noqa: E501
    """

    api_token: SecretStr
    """ API access token for HuggingFace."""

    llm: InferenceClient

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.llm = InferenceClient(
            self.model, api_key=self.api_token.get_secret_value()
        )

    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> str:
        """Run zero-shot classification on the given input.

        Args:
            prompt: The prompt to generate from.
            stop: Stop words to use when generating. Model output is cut
                off at the first occurrence of any of the stop substrings.
                If stop tokens are not supported consider raising
                NotImplementedError.
            run_manager: Callback manager for the run.
            **kwargs: Arbitrary additional keyword arguments. These are
                usually passed to the model provider API call.

        Returns:
            The model output as a string. Actual completions SHOULD NOT
            include the prompt.
        """
        if stop is not None:
            raise ValueError("stop kwargs are not permitted.")
        classification_result = self.llm.zero_shot_classification(
            text=prompt,
            candidate_labels=self.labels,
            multi_label=self.multi_label,
            hypothesis_template=self.hypothesis_template,
        )
        print(f"zero shot classification return: " f"{classification_result[0].label}")
        # we simply return the first result element (which has by
        # convention the highest probability (?))
        return classification_result[0].label

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Return a dictionary of identifying parameters."""
        return {
            # The model name allows users to specify custom token counting
            # rules in LLM monitoring applications (e.g., in LangSmith users
            # can provide per token pricing for their model and monitor
            # costs for the given LLM.)
            "model_name": self.model,
        }

    @property
    def _llm_type(self) -> str:
        """Get the type of language model used by this chat model.
        Used for logging purposes only."""
        return f"Zero Shot Classifier ({self.model})"


# TODO: Move serialization logic to separate manager class 
def _current_subdir(subdirectory_construction: str | None) -> str:
    if (
        subdirectory_construction is None
        or subdirectory_construction not in SubdirConstruction.value2formatcode().keys()
    ):
        return ""
    now = datetime.now()
    dateformat = SubdirConstruction.value2formatcode()[subdirectory_construction]
    return now.strftime(dateformat)


def _yaml_dump(
    dir_path: str,
    file_name: str,
    dict: Dict,
    local_serialization: bool,
    make_dirs: bool = False,
    key_name: str | None = None,
):
    # Set up YAML representers for enums
    def enum_representer(dumper, data):
        """YAML representer for enums - serialize as their value"""
        return dumper.represent_scalar("tag:yaml.org,2002:str", data.value)

    # Register representers
    yaml.add_representer(ToxicityAnswer, enum_representer)

    file_path = os.path.join(dir_path, file_name)
    if local_serialization:
        if make_dirs:
            os.makedirs(dir_path, exist_ok=True)
        with open(file_path, "w", encoding="utf-8") as f:
            yaml.dump(
                dict, f, allow_unicode=True, default_flow_style=False, encoding="utf-8"
            )
    else:
        # Strip 'datasets/' prefix if present to avoid duplication
        # (hf_base_path may include 'datasets/', but HfFileSystem path
        # should not duplicate it)
        clean_dir_path = dir_path
        if dir_path.startswith("datasets/"):
            clean_dir_path = dir_path[len("datasets/"):]

        file_path = os.path.join("hf://datasets", clean_dir_path, file_name).replace(
            "\\", "/"
        )
        if key_name:
            fs = HfFileSystem(token=os.environ[key_name])
        else:
            fs = HfFileSystem()
        if make_dirs:
            # Also strip prefix for makedirs
            clean_makedirs_path = dir_path
            if dir_path.startswith("datasets/"):
                clean_makedirs_path = dir_path[len("datasets/"):]
            fs.makedirs(clean_makedirs_path, exist_ok=True)
        # HfFileSystem requires binary mode
        with fs.open(file_path, "wb") as f:
            # yaml.dump to string first, then encode to bytes
            yaml_str = yaml.dump(dict, allow_unicode=True, default_flow_style=False)
            f.write(yaml_str.encode("utf-8"))


def _yaml_load(
    file_path: str, local_serialization: bool, key_name: str | None = None
) -> Dict:
    if local_serialization:
        with open(file_path, "r", encoding="utf-8") as f:
            ret_dict = yaml.safe_load(f)
    else:
        # Strip 'datasets/' prefix if present to avoid duplication
        clean_file_path = file_path
        if file_path.startswith("datasets/"):
            clean_file_path = file_path[len("datasets/"):]
        file_path = os.path.join("hf://datasets", clean_file_path).replace("\\", "/")
        if key_name:
            fs = HfFileSystem(token=os.environ[key_name])
        else:
            fs = HfFileSystem()
        # HfFileSystem requires binary mode, no encoding parameter
        with fs.open(file_path, "rb") as f:
            ret_dict = yaml.safe_load(f)
    return ret_dict


def _str_load(
    file_path: str, local_serialization: bool, key_name: str | None = None
) -> str:
    logger.info(f"Getting file {file_path}")
    if local_serialization:
        with open(file_path, encoding="utf-8") as f:
            # Read the contents of the file into a variable
            f_str = f.read()
    else:
        # Strip 'datasets/' prefix if present to avoid duplication
        clean_file_path = file_path
        if file_path.startswith("datasets/"):
            clean_file_path = file_path[len("datasets/"):]
        file_path = os.path.join("hf://datasets", clean_file_path).replace("\\", "/")
        if key_name:
            fs = HfFileSystem(token=os.environ[key_name])
        else:
            fs = HfFileSystem()
        # HfFileSystem requires binary mode, no encoding parameter
        with fs.open(file_path, "rb") as f:
            # Read the contents of the file as bytes and decode
            f_str = f.read().decode("utf-8")
    logger.info(f"Got file {file_path}")
    return f_str


def _str_dump(
    file_path: str,
    file_str: str,
    local_serialization: bool,
    key_name: str | None = None,
):
    if local_serialization:
        with open(file_path, "w", encoding="utf-8") as f:
            f.write(file_str)
    else:
        # Strip 'datasets/' prefix if present to avoid duplication
        clean_file_path = file_path
        if file_path.startswith("datasets/"):
            clean_file_path = file_path[len("datasets/"):]
        file_path = os.path.join("hf://datasets", clean_file_path).replace("\\", "/")
        if key_name:
            fs = HfFileSystem(token=os.environ[key_name])
        else:
            fs = HfFileSystem()
        # HfFileSystem requires binary mode
        with fs.open(file_path, "wb") as f:
            f.write(file_str.encode("utf-8"))


def update_feedback(
    pipeline_config: PipelineConfig,
    result: ToxicityDetectorResult,
    feedback_text: str | None = None,
    feedback_correctness: str | None = None,
    feedback_likert_content: Dict | None = None,
):

    result.feedback["feedback_text"] = feedback_text
    result.feedback["correctness"] = feedback_correctness
    if feedback_likert_content:
        for key, value in feedback_likert_content.items():
            result.feedback[key] = value

    save_result(result, pipeline_config)


def save_result(result: ToxicityDetectorResult, pipeline_config: PipelineConfig):
    subdirectory_path = _current_subdir(pipeline_config.subdirectory_construction)
    local_serialization = pipeline_config.local_serialization
    file_name = f"{result.request_id}.yaml"
    dir_path = os.path.join(
        pipeline_config.get_base_path(),
        pipeline_config.result_data_path,
        subdirectory_path,
    )
    _yaml_dump(
        dir_path,
        file_name,
        result.model_dump(),
        local_serialization,
        make_dirs=True,
        key_name=pipeline_config.hf_key_name,
    )


def get_openai_chat_model(
    base_url: str,
    model: str,
    api_key: SecretStr,
) -> ChatOpenAI:
    """Return ChatOpenAI model with given server URL, model, API key.

    Args:
    base_url (str): The URL of the inference server.
    model (str): Model (repo id).
    api_key (SecretStr): The API key for accessing the inference
        server.

    Returns:
    model (ChatOpenAI): An instance of the ChatOpenAI model.
    """
    chat_model = ChatOpenAI(base_url=base_url, model=model, api_key=api_key)

    return chat_model



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, Optional
import re

from operator import itemgetter

from langchain_core.runnables import (
    Runnable,
    RunnablePassthrough,
    RunnableParallel,
)
from langchain_core.language_models.base import BaseLanguageModel
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.output_parsers import StrOutputParser

from toxicity_detector.datamodels import ToxicityAnswer


class BaseChainBuilder(ABC):
    """Abstract Base Class for chain builders based on langchain"""

    # TODO: How to handle diverging model kwargs?
    # How to individually configure models?
    @classmethod
    @abstractmethod
    def build(cls, llms_dict: Dict[str, BaseLanguageModel], **kwargs) -> Runnable:
        """Abstract factory method to build an LLM chain.

        Args:
            llms_dict (Dict[str, BaseLanguageModel]): The models to be
                used (names, keys are given by convention).
            **kwargs: Additional parameters including prompts and model kwargs.

        Returns:
            Runnable: Chain
        """
        pass


class IdentifyToxicContentZeroShotChain(BaseChainBuilder):
    """Chain using chat model to explain/justify toxicity categorisation."""

    @classmethod
    def build(  # type: ignore[override]
        cls,
        llms_dict: Dict[str, BaseLanguageModel],
        explain_toxicity: list[tuple[str, str]],
        **model_kwargs,
    ) -> Runnable:
        """Simple chain based on zero-shot model for categorization.

        Uses a chat model for ex post justification of the categorization.

        Args:
            llms_dict (Dict[str, BaseLanguageModel]): A dict with one
                model of the form `{'zero_shot_model': the_zero_shot_model,
                'chat_model': the_chat_model}`
            explain_toxicity: List of (role, content) tuples for the prompt

        Returns:
            Runnable: Chain
        """
        main_chain = (
            RunnableParallel(
                passed=RunnablePassthrough(),
                zero_shot_classification=itemgetter("user_input")
                | llms_dict["zero_shot_model"].bind(**model_kwargs),
            )
            | (
                lambda inputs: {
                    "toxicity_value": (
                        "toxic"
                        if inputs["zero_shot_classification"]
                        == inputs["passed"]["labels"]["toxic"]
                        else "not toxic"
                    ),
                    "toxicity_explication": inputs["passed"]["toxicity_explication"],
                    "user_input": inputs["passed"]["user_input"],
                }
            )
            | ChatPromptTemplate.from_messages(explain_toxicity)
            | llms_dict["chat_model"].bind(**model_kwargs)
        )
        return main_chain


class MonoModelDetectToxicityChain(BaseChainBuilder):
    """Chain builder for detecting toxicity.

    Caveats:
    - This generated chain uses one model only. It is not possible to
      assign different models to the subtasks.

    Input:
    - toxicity_explication: An explication of the toxicity concept.
    - user_input: The text that should be analyzed.
    - general_questions: `Dict[str, str]` of the form
      `{'name': <name>, 'llm_description': <partial prompt>}`.
    - context_information: Context information about the text (may be
      empty).
    """

    @staticmethod
    def _parse_toxicity_response(response: str) -> Optional[ToxicityAnswer]:
        """Parse the toxicity response from the LLM.

        Handles variations in the response format including prefixes like
        "Answer:", "Antwort:", "The answer is:", etc.

        Args:
            response (str): The LLM response to parse

        Returns:
            Optional[ToxicityAnswer]:
                - ToxicityAnswer.TRUE if the text contains toxic content
                - ToxicityAnswer.FALSE if the text doesn't contain toxic content
                - ToxicityAnswer.UNCLEAR if the result is unclear
                - None if the response cannot be parsed
        """
        # Normalize the response: strip whitespace, remove quotes, lowercase
        normalized = response.strip().strip("\"'").lower()

        # Remove common prefixes (English and German)
        prefix_patterns = [
            r"^(answer|antwort|the answer is|die antwort ist|result|ergebnis)\s*:?\s*",
        ]
        for pattern in prefix_patterns:
            normalized = re.sub(pattern, "", normalized, flags=re.IGNORECASE)
        normalized = normalized.strip().strip("\"'")

        # Check for TRUE indicators (toxic content)
        true_indicators = ["true", "yes", "ja", "toxic", "toxisch", "wahr"]
        if normalized in true_indicators:
            return ToxicityAnswer.TRUE

        # Check for FALSE indicators (non-toxic content)
        false_indicators = [
            "false",
            "no",
            "nein",
            "not toxic",
            "non-toxic",
            "nicht toxisch",
            "falsch",
        ]
        if normalized in false_indicators:
            return ToxicityAnswer.FALSE

        # Check for UNCLEAR indicators
        unclear_indicators = [
            "unclear",
            "unklar",
            "unsure",
            "unknown",
            "unbekannt",
            "uncertain",
            "unsicher",
        ]
        if normalized in unclear_indicators:
            return ToxicityAnswer.UNCLEAR

        # If response contains these keywords, extract the meaning
        if any(keyword in normalized for keyword in ["true", "toxic"]):
            # Make sure it's not negated
            if not any(neg in normalized for neg in ["not", "non", "nicht", "kein"]):
                return ToxicityAnswer.TRUE

        if any(
            keyword in normalized
            for keyword in ["false", "not toxic", "non-toxic", "nicht toxisch"]
        ):
            return ToxicityAnswer.FALSE

        if any(keyword in normalized for keyword in ["unclear", "unklar", "unsure"]):
            return ToxicityAnswer.UNCLEAR

        # Default to None if we can't parse the response
        return None

    @classmethod
    def prompts(
        cls,
        preprocessing: list[tuple[str, str]],
        indicator_classification: list[tuple[str, str]],
        indicator_aggregation: list[tuple[str, str]],
        **kwargs: Any,
    ) -> Dict:
        """Generate formatted prompts for debugging/logging.

        Args:
            preprocessing: Preprocessing prompt template
            indicator_classification: Indicator classification
                prompt template
            indicator_aggregation: Aggregation prompt template
            **kwargs: Variables to format into the prompts

        Returns:
            Dict with formatted prompts
        """
        ret_dict = {
            "preprocessing": ChatPromptTemplate.from_messages(
                preprocessing, template_format="jinja2"
            )
            .format_prompt(**kwargs)
            .to_string(),
            "indicators": {
                indicator_key: ChatPromptTemplate.from_messages(
                    indicator_classification,
                    template_format="jinja2",
                )
                .format_prompt(
                    indicator_description=indicator["llm_description"],
                    indicator_name=indicator["name"],
                    **kwargs,
                )
                .to_string()
                for indicator_key, indicator in kwargs["indicators_dict"].items()
            },
            "aggregation": ChatPromptTemplate.from_messages(
                indicator_aggregation, template_format="jinja2"
            )
            .format_prompt(**kwargs)
            .to_string(),
        }
        return ret_dict

    @classmethod
    def build(  # type: ignore[override]
        cls,
        llms_dict: Dict[str, BaseLanguageModel],
        preprocessing: list[tuple[str, str]],
        indicator_classification: list[tuple[str, str]],
        indicator_aggregation: list[tuple[str, str]],
        formatting_prompt_msgs: list[tuple[str, str]],
        indicators_dict: Dict[str, Dict[str, str]] | None = None,
        **model_kwargs,
    ) -> Runnable:
        """Builds a chain for identifying toxicity.

        Args:
            llms_dict (Dict[str, BaseLanguageModel]): A dict with one
                model of the form `{'chat_model': the_chat_model}`
            preprocessing: List of (role, content) tuples for
                preprocessing prompt
            indicator_classification: List of (role, content) tuples
                for indicator prompt
            indicator_aggregation: List of (role, content) tuples for
                aggregation prompt
            formatting_prompt_msgs: List of (role, content) tuples for
                formatting prompt
            indicators_dict (Dict[str, Dict[str, str]]): A dict with all
                relevant indicators of the form
                `{<indicator_key>: {'name': <indicator name>,
                'llm_description': <partial prompt>},...}`

        Returns:
            Runnable: Chain
        """
        if indicators_dict is None:
            indicators_dict = {}
        llm = llms_dict["chat_model"]

        # sub chain: preprocessing: analysing text w.r.t.
        # general (relevant) properties
        general_props_chain = (
            ChatPromptTemplate.from_messages(preprocessing, template_format="jinja2")
            | llm.bind(**model_kwargs)
            | StrOutputParser()
        )

        # sub chain: indicator chain
        def indicator_chain(indicator_name: str, indicator_description: str):
            chain = (
                RunnablePassthrough.assign(
                    indicator_name=lambda x: indicator_name,
                    indicator_description=lambda x: indicator_description,
                )
                | ChatPromptTemplate.from_messages(
                    indicator_classification,
                    template_format="jinja2",
                )
                | llm.bind(**model_kwargs)
                | StrOutputParser()
            )
            return chain

        # sub chain: aggregation chain
        aggregation_chain = (
            ChatPromptTemplate.from_messages(
                indicator_aggregation, template_format="jinja2"
            )
            | llm.bind(**model_kwargs)
            | StrOutputParser()
        )

        main_chain = (
            # We add the indicators dict to the
            # General questions chain (preprocessing)
            RunnablePassthrough.assign(preprocessing_results=general_props_chain)
            # Branches: independent indicator chains
            | RunnablePassthrough.assign(
                indicator_analysis=RunnableParallel(
                    {
                        indicator_key: indicator_chain(
                            indicator["name"], indicator["llm_description"]
                        )
                        for indicator_key, indicator in indicators_dict.items()
                    }
                )
            )
            # Aggregation of preliminary results to an overall categorisation
            | RunnablePassthrough.assign(analysis_result=aggregation_chain)
            # Adding binary answer
            | RunnablePassthrough.assign(
                contains_toxicity=ChatPromptTemplate.from_messages(
                    formatting_prompt_msgs, template_format="jinja2"
                )
                # TODO: Perhaps, set temperature to 0
                | llm.bind(**model_kwargs)
                | StrOutputParser()
                | cls._parse_toxicity_response
            )
        )

        return main_chain


class IdentifyToxicContentChatChain(BaseChainBuilder):
    """Chain that uses a Chatmodel for the categorisation."""

    # for debuggin/loggin only:
    @classmethod
    def format_prompt(
        cls, identify_toxicity: list[tuple[str, str]], **kwargs: Any
    ) -> str:
        """Format prompt for debugging/logging.

        Args:
            identify_toxicity: List of (role, content) tuples for the prompt
            **kwargs: Variables to format into the prompt

        Returns:
            Formatted prompt as string
        """
        return (
            ChatPromptTemplate.from_messages(identify_toxicity)
            .format_prompt(**kwargs)
            .to_string()
        )

    @classmethod
    def build(  # type: ignore[override]
        cls,
        llms_dict: Dict[str, BaseLanguageModel],
        identify_toxicity: list[tuple[str, str]],
        **model_kwargs,
    ) -> Runnable:
        """Builds a simplistic chain for identifying toxicity.

        Args:
            llms_dict (Dict[str, BaseLanguageModel]): A dict with one
                model of the form `{'chat_model': the_chat_model}`
            identify_toxicity: List of (role, content) tuples for the prompt

        Returns:
            Runnable: Chain
        """
        main_chain = (
            ChatPromptTemplate.from_messages(identify_toxicity)
            | llms_dict["chat_model"].bind(**model_kwargs)
            | StrOutputParser()
        )

        return main_chain



from pydantic import BaseModel, Field, field_validator, model_validator
import enum
from typing import Dict, Self, Optional, Any
import yaml
from huggingface_hub import HfFileSystem
import os
from loguru import logger
from datetime import datetime
from functools import lru_cache
from importlib.resources import files

from toxicity_detector.datamodels import Toxicity, ToxicityType

# Minimum required pipeline config version
MIN_PIPELINE_CONFIG_VERSION = "v0.4"
# CONFIG VERSION HISTORY
# v0.1: initial version (yaml only)
# v0.3:
#   + new field: toxicity_examples_data_file
#   + new subfield: general_questions.name and general_questions.llm_description
#     (instead of putting the prompt directly under general_questions)
# v0.4: wrapping config with Pydantic (PipelineConfig)
#   + removing field: toxicity_examples_data_file (moved to AppConfig)
#   + new fields: local_serialization, hf_base_path, hf_key_name, local_base_path,
#     result_data_path, log_path, subdirectory_construction, env_file (moved from
#     the app config)
#   + renamed: `personalized_toxicity` -> `personalized_toxic_speech`
#     (!BREAKING CHANGE!)
# v0.5: moving remaing hard coded prompts to config
#   + only adding fields -> no breaking changes
#   + new fields: prompts.


@lru_cache(maxsize=1)
def _load_default_config_dict() -> Dict[str, Any]:
    """Load default configuration from YAML. Cached for performance."""
    try:
        # Use importlib.resources to access package data
        config_file = files("toxicity_detector.package_data").joinpath(
            "default_pipeline_config.yaml"
        )

        with config_file.open("r", encoding="utf-8") as f:
            return yaml.safe_load(f)
    except (FileNotFoundError, AttributeError) as e:
        logger.error(
            f"Failed to load default configuration from package data: {e}. "
            "Ensure package_data is properly installed."
        )
        raise


def _get_default_for_field(field_name: str) -> Any:
    """Get default value for a specific field from YAML."""
    config_dict = _load_default_config_dict()
    return config_dict.get(field_name)


def _get_default_toxicities() -> Dict[str, Toxicity]:
    """
    Get default toxicities from YAML and convert to Toxicity instances.

    This is needed because when using default_factory with a lambda that returns
    a dict, Pydantic doesn't perform validation/conversion to model instances.
    We explicitly convert the raw YAML dict to Toxicity instances here.
    """
    toxicities_dict = _get_default_for_field("toxicities")
    if not toxicities_dict:
        return {}

    # Convert each toxicity dict to a Toxicity model instance
    return {key: Toxicity(**value) for key, value in toxicities_dict.items()}


class SubdirConstruction(enum.Enum):
    DAILY = "daily"
    WEEKLY = "weekly"
    MONTHLY = "monthly"
    YEARLY = "yearly"

    @classmethod
    def value2formatcode(cls) -> Dict[str, str]:
        return {
            cls.DAILY.value: "%Y_%m_%d",
            cls.WEEKLY.value: "y%Y_w%W",
            cls.MONTHLY.value: "y%Y_m%m",
            cls.YEARLY.value: "y%Y",
        }


class PipelineConfig(BaseModel):
    local_serialization: bool = Field(
        default_factory=lambda: _get_default_for_field("local_serialization") or True
    )
    hf_base_path: str | None = Field(
        default_factory=lambda: _get_default_for_field("hf_base_path")
    )
    hf_key_name: str | None = Field(
        default_factory=lambda: _get_default_for_field("hf_key_name")
    )
    result_data_path: str = Field(
        default_factory=lambda: (
            _get_default_for_field("result_data_path") or "result_data"
        )
    )
    local_base_path: str | None = Field(
        default_factory=lambda: _get_default_for_field("local_base_path")
    )
    log_path: str = Field(
        default_factory=lambda: _get_default_for_field("log_path") or "logs"
    )
    # one of monthly, weekly, yearly, daily, None
    subdirectory_construction: Optional[str] = Field(
        default_factory=lambda: (_get_default_for_field("subdirectory_construction"))
    )
    toxicities: Dict[str, Toxicity] = Field(default_factory=_get_default_toxicities)
    config_version: str = Field(
        default_factory=lambda: _get_default_for_field("config_version")
    )
    used_chat_model: str
    description: str | None = Field(
        default_factory=lambda: _get_default_for_field("description")
    )
    system_prompt: str = Field(
        default_factory=lambda: _get_default_for_field("system_prompt")
        or (
            "You are a helpful assistant and an expert for the "
            "categorisation and annotation of texts.\n"
            "You read instructions carefully and follow them precisely.\n"
            "You give concise and clear answers."
        )
    )
    models: Dict[str, Dict[str, Any]] = Field(
        default_factory=lambda: _get_default_for_field("models") or {}
    )
    env_file: str | None = Field(
        default_factory=lambda: _get_default_for_field("env_file")
    )
    prompt_templates: Dict[str, list[Dict[str, str]]] = Field(
        default_factory=lambda: _get_default_for_field("prompt_templates") or {}
    )

    def __init__(self, **data):
        super().__init__(**data)
        self._configure_logging()

    def _configure_logging(self):
        """Configure loguru file handler for logging."""
        if self.local_serialization:
            log_dir_path = os.path.join(
                self.get_base_path(),
                self.log_path,
            )
            os.makedirs(log_dir_path, exist_ok=True)

            # Create log file name with current date
            now = datetime.now()
            log_file_name = now.strftime("toxicity_detector_log_%Y_%m_%d.log")
            log_file_path = os.path.join(log_dir_path, log_file_name)

            # Add file handler to loguru
            # Use rotation="1 day" to create a new file each day
            logger.add(
                log_file_path,
                rotation="1 day",
                retention="30 days",
                format=(
                    "{time:YYYY-MM-DD HH:mm:ss} | {level} | "
                    "{name}:{function}:{line} - {message}"
                ),
                level="INFO",
            )
            logger.info(f"Configured logging to file: {log_file_path}")

    @model_validator(mode="after")
    def load_env_file(self) -> Self:
        if self.env_file is None:
            logger.warning(
                "No environment file with API keys specified for Pipeline Config."
                " Please set 'env_file' to a valid path if you want "
                "to load environment variables from a file."
            )
        else:
            # check if the env file exists
            from os import path

            # env_file_path = path.join(
            #     self.get_base_path(),
            #     self.env_file,
            # )
            # env path should be absolute or relative to working dir
            # instead of relative to base path (since this might point to HF)
            env_file_path = self.env_file

            if not path.exists(env_file_path):
                err_msg = (
                    f"Environment file '{env_file_path}' does not exist. "
                    "Please provide a valid path to the environment file. "
                    "Or set it to None if you don't need it and set the "
                    "API keys in other ways as environment variables."
                )
                logger.warning(err_msg)
            else:
                # load the env file
                from dotenv import load_dotenv

                load_dotenv(env_file_path)
                logger.info(f"Loaded environment variables from '{env_file_path}'")
        return self

    @model_validator(mode="after")
    def validate_models(self) -> Self:
        allowed_values = set(self.models.keys())
        if len(allowed_values) == 0:
            raise ValueError("At least one model must be specified under models")
        elif self.used_chat_model not in allowed_values:
            raise ValueError(
                f"used_chat_model must be one of {allowed_values}, "
                f"got {self.used_chat_model}"
            )
        return self

    @field_validator("toxicities")
    def validate_toxicity_types(cls, v):
        allowed_values = ToxicityType._value2member_map_.keys()
        toxicity_types = v.keys()
        if not set(toxicity_types).issubset(set(allowed_values)):
            raise ValueError(
                f"Allowed toxicities are {set(allowed_values)}, "
                f"got {set(toxicity_types)}"
            )
        return v

    @field_validator("config_version")
    @classmethod
    def validate_config_version(cls, v):
        if v is None:
            raise ValueError(
                f"config_version must be specified. "
                f"Minimum required version is {MIN_PIPELINE_CONFIG_VERSION}"
            )

        # Extract version numbers from format "vX.Y" or "vX.Y.Z"
        def parse_version(version_str: str) -> tuple:
            # Remove 'v' prefix and split by '.'
            version_parts = version_str.lstrip("v").split(".")
            return tuple(int(part) for part in version_parts)

        try:
            current_version = parse_version(v)
            min_version = parse_version(MIN_PIPELINE_CONFIG_VERSION)

            if current_version < min_version:
                raise ValueError(
                    f"config_version {v} is below minimum required version "
                    f"{MIN_PIPELINE_CONFIG_VERSION}"
                )
        except (ValueError, AttributeError) as e:
            if "invalid literal" in str(e):
                raise ValueError(
                    f"Invalid config_version format: {v}. "
                    f"Expected format like 'v0.3' or 'v0.4.1'"
                )
            raise

        return v

    @field_validator("subdirectory_construction")
    @classmethod
    def validate_subdirectory_construction(cls, v):
        allowed_values = SubdirConstruction._value2member_map_.keys()
        if (v is not None) and (v not in allowed_values):
            raise ValueError(
                f"subdirectory_construction must be one "
                f"of {set(allowed_values)}, got {v}"
            )
        return v

    @model_validator(mode="after")
    def validate_serialization(self) -> Self:
        if self.local_serialization and self.local_base_path is None:
            raise ValueError(
                "local_base_path must be set if choosing local serialization"
            )
        elif not self.local_serialization and self.hf_base_path is None:
            raise ValueError(
                "hf_base_path must be set if not choosing local serialization"
            )
        return self

    def get_base_path(self) -> str:
        if self.local_serialization:
            assert self.local_base_path is not None
            return self.local_base_path
        else:
            assert self.hf_base_path is not None
            return self.hf_base_path

    def get_prompt_messages(self, prompt_name: str) -> list[tuple[str, str]]:
        """Get prompt messages as list of (role, content) tuples for LangChain.

        Args:
            prompt_name: Name of the prompt in the config

        Returns:
            List of (role, content) tuples ready for ChatPromptTemplate.from_messages

        Raises:
            ValueError: If prompt_name is not found in configuration
        """
        if prompt_name not in self.prompt_templates:
            raise ValueError(f"Prompt '{prompt_name}' not found in configuration")

        prompt_list = self.prompt_templates[prompt_name]
        return [(msg["role"], msg["content"]) for msg in prompt_list]

    @staticmethod
    def from_file(file_path: str) -> "PipelineConfig":
        with open(file_path, encoding="utf-8") as f:
            config = PipelineConfig(**yaml.safe_load(f.read()))
        return config


# TODO: mv defaults to PKG data YAML file and load them during init as defaults
# TODO: add all other UI texts here (to make them configurable)
class UITexts(BaseModel):
    trigger_warning: dict[str, str] = {
        "checkbox_label": (
            "Die Hinweise habe ich zur Kenntnis genommen und der Speicherung "
            "der Daten stimme ich zu."
        ),
        "message": (
            "# Benutzunghinweise (*!vorläufige Formulierung!*)\n\n"
            "+ **Triggerwarnung:** Die Toxizitätdetektorapp enthält in Form "
            "von Beispielen Inhalte, die anstößig oder beunruhigend sein können. "
            "Alle Materialien dienen der Unterstützung von Forschungsarbeiten zur "
            "Verbesserung der Methoden zur Erkennung von Toxizität. Die enthaltenen "
            "Beispiele für Toxizität geben insbesondere nicht wieder, wie die Autoren "
            "über bestimmte Identitätsgruppen bzw. Personen denken. Die Beispiele "
            "stammen aus dem Korpus ...\n"
            "+ **Datenerhebung:** Die eingegebenen Textbeispiele und generierten "
            "Kategorisierungen werden für Forschungszwecke gespeichert und benutzt "
            "um die Performance von Modelle zu steigern. Darüberhinaus werden keine "
            "Daten gesammelt, insbesondere keine personenbezogenen Daten (sofern keine "
            "personenbezogenen Daten in das Textfeld eingetragen werden)\n"
        ),
    }
    app_head: str = (
        "# 📣 Detektor für toxische Sprache\n"
        "In dieser Demoapp kannst Du ausprobieren, wie gut Large Language Models "
        "Toxizität detektieren können.\n"
    )


class AppConfig(BaseModel):
    developer_mode: bool = False
    pipeline_config_version: str | None = None
    toxicity_examples_data_file: str | None = None
    local_serialization: bool | None = None
    hf_base_path: str | None = None
    hf_key_name: str | None = None
    local_base_path: str | None = None
    config_path: str = "configs"
    env_file: str | None = None
    pipeline_config_file: str
    force_agreement: bool = False

    # TODO: mv defaults to PKG data YAML file and load them during init as defaults
    feedback: Dict[str, Any] = {
        "likert_scale": {
            "absolutely_correct": "Stimmt absolut",
            "correct": "Stimmt",
            "suspension": "Unklar",
            "incorrect": "Stimmt nicht",
            "absolutely_incorrect": "Stimmt überhaupt nicht",
        }
    }
    ui_texts: UITexts = UITexts()

    def load_env_file(self):
        if self.env_file is None:
            logger.warning(
                "No environment file with API keys specified for App Config."
                " Please set 'env_file' to a valid path if you want "
                "to load environment variables from a file."
            )
        else:
            # check if the env file exists
            from os import path

            if not path.exists(self.env_file):
                err_msg = (
                    f"Environment file '{self.env_file}' does not exist. "
                    "Please provide a valid path to the environment file. "
                    "Or set it to None if you don't need it and set the "
                    "API keys in other ways as environment variables."
                )
                logger.warning(err_msg)
            else:
                # load the env file
                from dotenv import load_dotenv

                load_dotenv(self.env_file)
            logger.info(f"Loaded environment variables from '{self.env_file}'")

    @model_validator(mode="after")
    def _load_env_file(self) -> Self:
        self.load_env_file()
        return self

    # def get_pipeline_config_path(self) -> str:
    #     if self.local_pipeline_config:
    #         base_path = self.local_pipeline_config_base_path
    #     else:
    #         base_path = self.hf_pipeline_config_base_path
    #     assert base_path is not None
    #     return base_path

    # def get_default_pipeline_config(self) -> PipelineConfig:
    #     if self.local_pipeline_config:
    #         file_path = os.path.join(
    #             self.get_pipeline_config_path(), self.default_pipeline_config_file
    #         )
    #         return PipelineConfig.from_file(file_path)
    #     else:
    #         if self.pipeline_config_key_name:
    #             fs = HfFileSystem(token=os.environ[self.pipeline_config_key_name])
    #         else:
    #             fs = HfFileSystem()
    #         file_path = os.path.join(
    #             "hf://datasets",
    #             self.get_pipeline_config_path(),
    #             self.default_pipeline_config_file,
    #         ).replace("\\", "/")
    #         with fs._open(file_path, "rt", encoding="utf_8") as file:
    #             return PipelineConfig(**yaml.safe_load(file))

    @staticmethod
    def from_file(file_path: str) -> "AppConfig":
        with open(file_path, encoding="utf_8") as f:
            config = AppConfig(**yaml.safe_load(f))
        return config



"""Detection tab for the Gradio app."""

import gradio as gr
import pandas as pd
from typing import Dict, Tuple
from random import randrange

from toxicity_detector import detect_toxicity, update_feedback
from toxicity_detector.config import PipelineConfig
from toxicity_detector.result import ToxicityDetectorResult
from toxicity_detector.app.app_config_loader import _app_config, _config_manager


def _tasks(pipeline_config: PipelineConfig, toxicity_type) -> list[str]:
    """Get list of task names for a toxicity type."""
    task_names = []
    task_groups = pipeline_config.toxicities[toxicity_type].tasks.keys()
    for task_group in task_groups:
        task_names.extend(
            list(pipeline_config.toxicities[toxicity_type].tasks[task_group].keys())
        )
    return task_names


def _load_toxicity_example_data() -> pd.DataFrame | None:
    """Load toxicity example data if configured."""
    if (data_file := _app_config().toxicity_examples_data_file) is not None:
        return _config_manager().load_toxicity_example_data(data_file)
    else:
        return None

# TODO: Use decorators consistently for event handling.

def create_detection_tab(
    tw_approved: gr.State,
    tw_approved_value: bool,
    pipeline_config_state: gr.State,
    result_state: gr.State,
    feedback_interactive_st: gr.State,
    output_dirty_st: gr.State,
    feedback_likert_content_st: gr.State,
    toxicity_example_data_st: gr.State,
    user_input_source_st: gr.State,
) -> gr.Tab:
    """Create the toxicity detection tab.

    Args:
        tw_approved: State for trigger warning approval
        tw_approved_value: Boolean value indicating if trigger warning was approved
        pipeline_config_state: State for pipeline configuration
        result_state: State for detection results
        feedback_interactive_st: State for feedback UI interactivity
        output_dirty_st: State for output dirty flag
        feedback_likert_content_st: State for feedback content
        toxicity_example_data_st: State for example data
        user_input_source_st: State for input source string

    Returns:
        The detection tab component
    """
    with gr.Tab(
        label="Toxizitätsdetektor",
        id="detector_tab",
        visible=tw_approved_value,
    ) as detector_tab:
        with gr.Row():
            with gr.Column(scale=1, min_width=300):
                init_toxicity_key = list(pipeline_config_state.value.toxicities.keys())[
                    0
                ]
                radio_toxicitiy_type = gr.Radio(
                    [
                        (value.title, key)
                        for (
                            key,
                            value,
                        ) in pipeline_config_state.value.toxicities.items()
                    ],
                    value=init_toxicity_key,
                    label="Toxizitätsdefinition",
                    info=("Welche Art von Toxizität soll " "detektiert werden?"),
                )
                with gr.Accordion("Definition der gewählten Toxizitätsart:"):
                    md_toxicity_description = gr.Markdown(
                        pipeline_config_state.value.toxicities[
                            init_toxicity_key
                        ].user_description
                    )

                user_input_tb = gr.Textbox(
                    label="Texteingabe",
                    info="Eingabe des zu kategorisierenden Textes.",
                    lines=6,
                )
                random_example_btn = gr.Button(
                    "Zufälliges Beispiel",
                    interactive=toxicity_example_data_st.value is not None,
                )
                with gr.Accordion(label="Kontextinfo (kann leer bleiben)", open=False):
                    context_tb = gr.Textbox(
                        lines=6,
                        container=False,
                    )
                categorize_btn = gr.Button("Detect Toxicity")

            with gr.Column(scale=2, min_width=300):
                with gr.Accordion(
                    label="Zwischenergebnisse der Pipeline (developer mode)",
                    visible=(True if _app_config().developer_mode else False),
                    open=False,
                ):
                    general_questions_output_tb = gr.Textbox(
                        label=("General questions output/preprocessing "),
                        visible=(True if _app_config().developer_mode else False),
                        lines=10,
                    )
                    indicators_output_tb = gr.Textbox(
                        label="Indicator analysis output (developer mode)",
                        visible=(True if _app_config().developer_mode else False),
                        lines=10,
                    )
                with gr.Accordion(
                    label="Ergebnis der Toxizitätskategorisierung", open=True
                ):
                    output_text_box = gr.Textbox(
                        label="Kategorisierung der Eingabe durch den Detektor",
                        lines=10,
                    )
                    output_label_text_box = gr.Textbox(
                        label="Label der Kategorisierung",
                        info="true/false/unclear ('unclear' = unsicher/unklar/nicht eindeutig)",
                    )
                feedback_radio = gr.Radio(
                    [
                        (value, key)
                        for (key, value) in _app_config()
                        .feedback["likert_scale"]
                        .items()
                    ],
                    label="Korrektheit der Kategorisierung",
                    info=(
                        "Stimmt die Kategorisierung des Detekors? "
                        "(Bist Du dir selbst unsicher, ob die Eingabe "
                        "toxischen Inhalt enthält, kreuze 'unklar' an.)"
                    ),
                    interactive=False,
                )
                feedback_textbox = gr.Textbox(
                    label="Feedback:",
                    info=(
                        "Hier kannst Du ausführliches Feedback zur "
                        "Kategoriesung des Textes durch den Detektor "
                        "eingeben."
                    ),
                    interactive=False,
                )
                with gr.Accordion(
                    "Taskspecific feedback (developer mode)",
                    visible=(True if _app_config().developer_mode else False),
                ):

                    @gr.render(
                        inputs=[
                            radio_toxicitiy_type,
                            feedback_interactive_st,
                        ]
                    )
                    def show_indicator_feedback_radios(
                        toxicity_type: str,
                        interactive: bool,
                    ):
                        for task in _tasks(pipeline_config_state.value, toxicity_type):
                            radio = gr.Radio(
                                [
                                    (value, key)
                                    for (
                                        key,
                                        value,
                                    ) in _app_config()
                                    .feedback["likert_scale"]
                                    .items()
                                ],
                                label=(
                                    f"Korrektheit der Antwort " f"(Indikator: {task})"
                                ),
                                info=(
                                    "Stimmt die Antwort/Beschreibung des "
                                    "Detekors? (Bist Du dir selbst "
                                    "unsicher, was eine korrekt Antwort "
                                    "ist, kreuze 'unklar' an.)"
                                ),
                                interactive=interactive,
                                value=None,
                            )

                            def update_indicator_feedback(
                                task: str,
                                indicator_feedback: str,
                                feedback_likert_content: Dict,
                            ):
                                if indicator_feedback:
                                    feedback_likert_content[task] = indicator_feedback
                                return feedback_likert_content

                            # event listener for the radio button
                            # (to update the feedback content)
                            radio.change(
                                lambda indicator_feedback, feedback_likert_content, task=task: update_indicator_feedback(  # noqa: E501
                                    task,
                                    indicator_feedback,
                                    feedback_likert_content,
                                ),
                                [radio, feedback_likert_content_st],
                                [feedback_likert_content_st],
                            )

                feedback_btn = gr.Button(
                    "Feedback speichern/aktualisieren", interactive=False
                )

        # EVENT LISTENER/LOGIC FOR DETECTION TAB
        def random_input_example(
            toxicity_example_data: pd.DataFrame | None,
        ) -> Tuple[str, str]:
            if toxicity_example_data is None or len(toxicity_example_data) == 0:
                return ("", "")
            example = toxicity_example_data.loc[
                randrange(len(toxicity_example_data)), :
            ]
            return (str(example["text"]), str(example["source"]))

        random_example_btn.click(
            random_input_example,
            toxicity_example_data_st,
            [user_input_tb, user_input_source_st],
        )
        # set output dirty when changing input
        user_input_tb.change(lambda: True, None, output_dirty_st)
        # set input source string if user edits the input
        user_input_tb.input(
            lambda: "kideku_toxicity_detector", None, user_input_source_st
        )

        # if changed to dirty, we clear the output textboxes and
        # deactivate the feedback ui
        output_dirty_st.change(
            lambda dirty: (
                (
                    gr.Textbox(interactive=False, value="")
                    if dirty
                    else gr.Textbox(interactive=False)
                ),
                (
                    gr.Textbox(interactive=False, value="")
                    if dirty
                    else gr.Textbox(interactive=False)
                ),
                (
                    gr.Textbox(interactive=False, value="")
                    if dirty
                    else gr.Textbox(interactive=False)
                ),
                (
                    gr.Textbox(interactive=False, value="")
                    if dirty
                    else gr.Textbox(interactive=False)
                ),
                not dirty,  # interactive feedback ui
                dict(),  # feedback content
            ),
            output_dirty_st,
            [
                output_text_box,
                output_label_text_box,
                general_questions_output_tb,
                indicators_output_tb,
                feedback_interactive_st,
                feedback_likert_content_st,
            ],
        )
        # de-/activation of feedback ui
        feedback_interactive_st.change(
            lambda interactive: (
                gr.Radio(interactive=interactive, value=None),
                gr.Textbox(interactive=interactive, value=""),
                gr.Button(interactive=interactive),
            ),
            feedback_interactive_st,
            [feedback_radio, feedback_textbox, feedback_btn],
        )

        # Detection button
        def detect_toxicity_wrapper(
            input_text: str,
            user_input_source: str,
            toxicity_type: str,
            context_info: str,
            pipeline_config: PipelineConfig,
        ):

            result = detect_toxicity(
                input_text=input_text,
                user_input_source=user_input_source,
                toxicity_type=toxicity_type,
                context_info=context_info,
                pipeline_config=pipeline_config,
            )

            indicator_result = result.answer["indicator_analysis"]
            # indicator analysis as one string for the ouput
            indicator_analysis_str = "".join(
                [
                    "".join([key, ": ", value, "\n\n"])
                    for key, value in indicator_result.items()
                ]
            )

            return (
                result.answer["analysis_result"],
                result.answer["contains_toxicity"].value,
                result.answer[
                    "preprocessing_results"
                ],  # ouput for text field (dev mode)
                indicator_analysis_str,  # output for textfield (dev mode)
                # feedback ui interactive via `feedback_interactive_st`
                True,
                dict(),  # feedback content
                False,  # output dirty
                result,
            )

        categorize_btn.click(
            fn=detect_toxicity_wrapper,
            inputs=[
                user_input_tb,
                user_input_source_st,
                radio_toxicitiy_type,
                context_tb,
                pipeline_config_state,
            ],
            outputs=[
                output_text_box,
                output_label_text_box,
                general_questions_output_tb,
                indicators_output_tb,
                feedback_interactive_st,
                feedback_likert_content_st,
                output_dirty_st,
                result_state,
            ],
        )
        # Changing toxicity type: -> update description
        # and set output uis to dirty
        radio_toxicitiy_type.change(
            lambda toxicity_type, pipeline_config: (
                pipeline_config.toxicities[toxicity_type].user_description,
                True,
            ),
            [radio_toxicitiy_type, pipeline_config_state],
            [md_toxicity_description, output_dirty_st],
        )

        # Saving feedback
        feedback_btn.click(
            lambda v, w, x, y, z: update_feedback(v, w, x, y, z),
            [
                pipeline_config_state,
                result_state,
                feedback_textbox,
                feedback_radio,
                feedback_likert_content_st,
            ],
            None,
        )

        # UPDATE UI ELEMENTS IF MODEL_CONFIG CHANGES
        # update toxicity description,
        # and set output ui elements to dirty
        def update_on_config_change(
            toxicity_type: str, pipeline_config: PipelineConfig
        ):
            return (
                pipeline_config.toxicities[toxicity_type].user_description,
                ToxicityDetectorResult(),
                True,  # set output ui elements to dirty
            )

        pipeline_config_state.change(
            update_on_config_change,
            [radio_toxicitiy_type, pipeline_config_state],
            [
                md_toxicity_description,
                result_state,
                output_dirty_st,
            ],
            show_progress="hidden",
        )

    return detector_tab



"""ConfigManager for handling pipeline configuration file operations."""

import os
import yaml
from glob import glob
from typing import Dict, List, Optional
from pathlib import Path
from huggingface_hub import HfFileSystem
from loguru import logger
import pandas as pd

from toxicity_detector.config import PipelineConfig


class ConfigManager:
    """Handles all configuration file operations (local and HuggingFace)."""

    def __init__(
        self,
        local_serialization: bool,
        local_base_path: str | None,
        hf_base_path: str | None,
        hf_key_name: str | None,
        config_path: str,
        default_pipeline_config_file: str,
        pipeline_config_version: str | None = None
    ):
        """Initialize ConfigManager.

        Args:
            local_serialization: If True, use local filesystem; if False,
                use HuggingFace
            local_base_path: Base path for local serialization
            hf_base_path: Base path for HuggingFace serialization (e.g., 'user/repo')
            hf_key_name: Environment variable name containing HF API token
            config_path: Relative path to config files from base_path
            default_pipeline_config_file: Default pipeline config filename
        """
        self.local_serialization = local_serialization
        # Normalize paths to avoid issues with ".", ".." etc.
        self.local_base_path = self._normalize_local_base_path(local_base_path)
        self.hf_base_path = self._normalize_hf_base_path(hf_base_path)
        self.hf_key_name = hf_key_name
        self.config_path = self._normalize_config_path(config_path)
        self.default_pipeline_config_file = default_pipeline_config_file
        self.pipeline_config_version = pipeline_config_version

        self._fs: Optional[HfFileSystem] = None

        # Validate configuration parameters
        self._validate_paths()
        logger.debug(
            "ConfigManager initialized with parameters: "
            f"local_serialization={self.local_serialization}, "
            f"local_base_path={self.local_base_path}, "
            f"hf_base_path={self.hf_base_path}, "
            f"hf_key_name={self.hf_key_name}, "
            f"config_path={self.config_path}, "
            f"default_pipeline_config_file={self.default_pipeline_config_file}, "
            f"pipeline_config_version={self.pipeline_config_version}"
        )

    def _normalize_local_base_path(self, path: str | None) -> str | None:
        """Normalize local base path to resolve '.', '..' etc.

        Args:
            path: Raw local base path that may contain relative components

        Returns:
            Normalized path or None if input was None
        """
        if path is None:
            return None

        # Normalize the path (resolves '.', '..')
        normalized = os.path.normpath(path)

        # Convert to forward slashes for consistency
        normalized = normalized.replace("\\", "/")

        # Warn if path contains '..' as it may be unexpected
        if ".." in path:
            logger.warning(
                f"local_base_path '{path}' contains '..' which resolves to: '{normalized}'. "
                f"Ensure this is the intended directory."
            )

        logger.debug(f"Normalized local_base_path from '{path}' to '{normalized}'")
        return normalized

    def _normalize_hf_base_path(self, path: str | None) -> str | None:
        """Normalize HuggingFace base path to resolve '.', '..' etc.

        Args:
            path: Raw HF base path that may contain relative components

        Returns:
            Normalized path or None if input was None
        """
        if path is None:
            return None

        # First normalize the path structure
        normalized = os.path.normpath(path)

        # Convert to forward slashes (required for HF)
        normalized = normalized.replace("\\", "/")

        # Remove leading './' if present
        if normalized.startswith("./"):
            normalized = normalized[2:]

        # Warn if path contains '..' - this is problematic for HF paths
        if ".." in path:
            logger.warning(
                f"hf_base_path '{path}' contains '..' which is resolved to: '{normalized}'. "
                f"This may not work as expected with HuggingFace repositories. "
                f"Consider using absolute repository paths like 'user/repo'."
            )

        logger.debug(f"Normalized hf_base_path from '{path}' to '{normalized}'")
        return normalized

    def _normalize_config_path(self, config_path: str) -> str:
        """Normalize config path to remove '.', '..' and ensure clean path.

        Args:
            config_path: Raw config path that may contain '.' or '..'

        Returns:
            Normalized path safe for both local and HF use
        """
        # Normalize the path (resolves '.', '..')
        normalized = os.path.normpath(config_path)

        # Convert to forward slashes for consistency (works on all platforms)
        normalized = normalized.replace("\\", "/")

        # Remove leading './' if present
        if normalized.startswith("./"):
            normalized = normalized[2:]

        # Warn if path tries to escape with '..'
        if normalized.startswith("..") or "/.." in normalized:
            logger.warning(
                f"config_path '{config_path}' contains '..' which may cause "
                f"unexpected behavior, especially with HuggingFace paths. "
                f"Normalized to: '{normalized}'"
            )

        # Special case: if just ".", treat as empty string (current directory)
        if normalized == ".":
            normalized = ""

        logger.debug(f"Normalized config_path from '{config_path}' to '{normalized}'")
        return normalized

    def _validate_paths(self):
        """Validate configuration paths based on serialization type."""
        if self.local_serialization:
            if not self.local_base_path:
                raise ValueError(
                    "local_base_path parameter must be set when "
                    "local_serialization is True."
                )
            if self.hf_base_path:
                logger.warning(
                    "hf_base_path parameter will be ignored since "
                    "local_serialization is True."
                )
        else:
            if not self.hf_base_path:
                raise ValueError(
                    "hf_base_path parameter must be set when "
                    "local_serialization is False."
                )
            if not self.hf_key_name or len(self.hf_key_name.strip()) == 0:
                logger.warning(
                    "hf_key_name not set for accessing HuggingFace path "
                    "Authentication may fail if the repository is private."
                )
            if self.local_base_path:
                logger.warning(
                    "local_base_path parameter will be ignored since "
                    "local_serialization is False."
                )

    @property
    def fs(self) -> HfFileSystem:
        """Lazy-load HuggingFace filesystem."""
        if not self.local_serialization and self._fs is None:
            token = os.environ.get(self.hf_key_name) if self.hf_key_name else None
            self._fs = HfFileSystem(token=token)
        return self._fs

    def _normalize_hf_path(self, path: str) -> str:
        """Remove 'datasets/' prefix if present to avoid duplication."""
        return path[len("datasets/") :] if path.startswith("datasets/") else path

    def _get_config_base_path(self) -> str:
        """Get the base path where config files are located."""
        if self.local_serialization:
            return os.path.join(
                self.local_base_path,  # type: ignore (checked by _validate_paths)
                self.config_path,  # type: ignore (checked by _validate_paths)
            )
        else:
            base = self._normalize_hf_path(self.hf_base_path)  # type: ignore
            # (checked by _validate_paths)
            return os.path.join(base, self.config_path).replace("\\", "/")

    def _get_full_path(self, file_name: str, for_hf_api: bool = False) -> str:
        """Get full path for a config file.

        Args:
            file_name: Name of the config file
            for_hf_api: If True, prepend 'hf://datasets/' for HF API calls

        Returns:
            Full path to the file
        """
        if self.local_serialization:
            return os.path.join(self._get_config_base_path(), file_name)
        else:
            path = os.path.join(self._get_config_base_path(), file_name).replace(
                "\\", "/"
            )
            return f"hf://datasets/{path}" if for_hf_api else path

    def load_yaml(self, file_name: str) -> Dict:
        """Load YAML file from local or HF storage.

        Args:
            file_name: Name of the YAML file to load

        Returns:
            Dictionary with loaded YAML content
        """
        logger.info(f"Loading YAML file: {file_name}")

        if self.local_serialization:
            path = self._get_full_path(file_name)
            with open(path, "r", encoding="utf-8") as f:
                return yaml.safe_load(f)
        else:
            path = self._get_full_path(file_name, for_hf_api=True)
            with self.fs.open(path, "rb") as f:
                return yaml.safe_load(f)

    def save_yaml(self, file_name: str, data: Dict, make_dirs: bool = False):
        """Save YAML file to local or HF storage.

        Args:
            file_name: Name of the YAML file to save
            data: Dictionary to save as YAML
            make_dirs: If True, create directories if they don't exist
        """
        logger.info(f"Saving YAML file: {file_name}")

        if self.local_serialization:
            path = self._get_full_path(file_name)
            if make_dirs:
                os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, "w", encoding="utf-8") as f:
                yaml.dump(data, f, allow_unicode=True, default_flow_style=False)
        else:
            if make_dirs:
                dir_path = self._get_config_base_path()
                self.fs.makedirs(f"hf://datasets/{dir_path}", exist_ok=True)
            path = self._get_full_path(file_name, for_hf_api=True)
            with self.fs.open(path, "wb") as f:
                yaml_str = yaml.dump(data, allow_unicode=True, default_flow_style=False)
                f.write(yaml_str.encode("utf-8"))

    def load_string(self, file_name: str) -> str:
        """Load text file as string.

        Args:
            file_name: Name of the file to load

        Returns:
            File contents as string
        """
        logger.info(f"Loading string file: {file_name}")

        if self.local_serialization:
            path = self._get_full_path(file_name)
            with open(path, "r", encoding="utf-8") as f:
                return f.read()
        else:
            path = self._get_full_path(file_name, for_hf_api=True)
            with self.fs.open(path, "rb") as f:
                return f.read().decode("utf-8")

    def get_pipeline_config_as_string(self, file_name: str | None = None) -> str:
        """Get pipeline config as string.

        Args:
            file_name: Config filename. If None, loads default config.

        Returns:
            Config file contents as string
        """
        file_name = file_name or self.default_pipeline_config_file
        logger.info(f"Loading pipeline config as string: {file_name}")
        return self.load_string(file_name)

    def save_string(self, file_name: str, content: str):
        """Save string to text file.

        Args:
            file_name: Name of the file to save
            content: String content to save
        """
        logger.info(f"Saving string file: {file_name}")

        if self.local_serialization:
            path = self._get_full_path(file_name)
            with open(path, "w", encoding="utf-8") as f:
                f.write(content)
        else:
            path = self._get_full_path(file_name, for_hf_api=True)
            with self.fs.open(path, "wb") as f:
                f.write(content.encode("utf-8"))

    def file_exists(self, file_name: str) -> bool:
        """Check if file exists.

        Args:
            file_name: Name of the file to check

        Returns:
            True if file exists, False otherwise
        """
        if self.local_serialization:
            path = self._get_full_path(file_name)
            return os.path.isfile(path)
        else:
            path = self._get_full_path(file_name, for_hf_api=True)
            return self.fs.exists(path)

    def list_config_files(self) -> List[str]:
        """List all pipeline config files, optionally filtered by version.

        Args:
            config_version: If provided, only return configs matching this version

        Returns:
            List of config filenames (without path)
        """
        logger.info(
            f"Listing config files (version filter: {self.pipeline_config_version})"
        )
        config_files = []

        if self.local_serialization:
            base_path = self._get_config_base_path()
            pattern = os.path.join(base_path, "*.yaml")

            for path in glob(pattern):
                try:
                    with open(path, "r", encoding="utf-8") as f:
                        data = yaml.safe_load(f)
                        # If no version filter, or version matches, add to list
                        if (
                            self.pipeline_config_version is None
                            or (
                                data.get("config_version") ==
                                self.pipeline_config_version
                            )
                        ):
                            config_files.append(os.path.basename(path))
                except Exception as e:
                    logger.warning(f"Failed to read {path}: {e}")
        else:
            base_path = self._get_config_base_path()
            pattern = f"hf://datasets/{base_path}/*.yaml"

            try:
                for path in self.fs.glob(pattern):
                    try:
                        with self.fs.open(path, "rb") as f:
                            data = yaml.safe_load(f)
                            # If no version filter, or version matches, add to list
                            if (
                                self.pipeline_config_version is None
                                or (
                                    data.get("config_version") ==
                                    self.pipeline_config_version
                                )
                            ):
                                config_files.append(os.path.basename(path))
                    except Exception as e:
                        logger.warning(f"Failed to read {path}: {e}")
            except Exception as e:
                logger.error(f"Failed to list files with pattern {pattern}: {e}")

        logger.info(f"Found {len(config_files)} config files")
        return sorted(config_files)

    def load_pipeline_config(self, file_name: str | None = None) -> PipelineConfig:
        """Load pipeline config from file.

        Args:
            file_name: Config filename. If None, loads default config.

        Returns:
            Loaded PipelineConfig object
        """
        file_name = file_name or self.default_pipeline_config_file
        logger.info(f"Loading pipeline config: {file_name}")

        config_dict = self.load_yaml(file_name)
        config = PipelineConfig(**config_dict)
        if (
            self.pipeline_config_version and
            config.config_version != self.pipeline_config_version
        ):
            logger.warning(
                f"Loaded pipeline config '{file_name}' has config_version "
                f"'{config.config_version}' which does not match the expected "
                f"version '{self.pipeline_config_version}'."
            )
        return config

    def get_default_pipeline_config(self) -> PipelineConfig:
        """Load the default pipeline config.

        Returns:
            Default PipelineConfig object
        """
        return self.load_pipeline_config()

    def save_pipeline_config(self, pipeline_config: PipelineConfig, file_name: str):
        """Save pipeline config to file.

        Args:
            pipeline_config: PipelineConfig object to save
            file_name: Filename to save to
        """
        logger.info(f"Saving pipeline config: {file_name}")
        self.save_yaml(file_name, pipeline_config.model_dump())

    def get_config_path(self) -> str:
        """Get the full config path (for backward compatibility).

        Returns:
            Full path to config directory
        """
        return self._get_config_base_path()

    def load_toxicity_example_data(self, data_file_path: str) -> pd.DataFrame:
        """Load toxicity example data from CSV file.

        Args:
            data_file_path: Relative path to CSV file from base_path

        Returns:
            DataFrame with 'text' and 'source' columns
        """
        logger.info(f"Loading toxicity example data from {data_file_path}")

        if self.local_serialization:
            full_path = os.path.join(
                self.local_base_path,  # type: ignore (checked by _validate_paths)
                data_file_path,
            )
            with open(full_path, "r", encoding="utf-8") as f:
                example_data_df = pd.read_csv(f)
        else:
            base = self._normalize_hf_path(self.hf_base_path)  # type: ignore
            # (checked by _validate_paths)
            hf_path = f"hf://datasets/{base}/{data_file_path}".replace("\\", "/")

            token = os.environ.get(self.hf_key_name) if self.hf_key_name else None
            fs = HfFileSystem(token=token)

            with fs.open(hf_path, "rb") as f:
                example_data_df = pd.read_csv(f)

        logger.info("Loading toxicity example data done.")
        return pd.DataFrame(example_data_df[["text", "source"]])



"""Agreement tab for the Gradio app."""

import gradio as gr

from toxicity_detector.app.app_config_loader import _app_config


def create_agreement_tab(
    tw_approved: gr.State,
    detector_tab: gr.Tab,
    config_tab: gr.Tab,
    tabs: gr.Tabs,
) -> gr.Tab:
    """Create the user agreement/trigger warning tab.
    
    Args:
        tw_approved: State for trigger warning approval
        detector_tab: The detector tab component
        config_tab: The config tab component
        tabs: The tabs container
        
    Returns:
        The agreement tab component
    """
    with gr.Tab(
        label="Benutzungshinweise",
        id="tw_tab",
        visible=_app_config().force_agreement,
    ) as tw_tab:
        gr.Markdown(_app_config().ui_texts.trigger_warning["message"])
        tw_checkbox = gr.Checkbox(
            label=_app_config().ui_texts.trigger_warning["checkbox_label"]
        )
        tw_checkbox.input(
            lambda x: (
                x,
                gr.Checkbox(interactive=False),
                gr.Tab(visible=x),
                gr.Tab(visible=x and _app_config().developer_mode),
                gr.Tabs(selected="detector_tab" if x else "tw_tab"),
            ),
            tw_checkbox,
            [tw_approved, tw_checkbox, detector_tab, config_tab, tabs],
        )
    
    return tw_tab



"""Main Gradio app module for toxicity detection."""

import gradio as gr
from loguru import logger

from toxicity_detector.result import ToxicityDetectorResult
from toxicity_detector.app.app_config_loader import (
    init_app_config,
    _app_config,
    _config_manager,
)
from toxicity_detector.app.detection_tab import (
    create_detection_tab,
    _load_toxicity_example_data,
)
from toxicity_detector.app.config_tab import create_config_tab
from toxicity_detector.app.agreement_tab import create_agreement_tab


def create_demo() -> gr.Blocks:
    """Create and return the Gradio demo interface."""
    with gr.Blocks(title="Chatbot Detektor für toxische Sprache") as demo:

        gr.Markdown(_app_config().ui_texts.app_head)

        # TODO: Refactor app to use a UI state class
        # Initialize state variables
        tw_approved_value = False if _app_config().force_agreement else True
        tw_approved = gr.State(tw_approved_value)
        result_state = gr.State(ToxicityDetectorResult())
        pipeline_config_state = gr.State(
            _config_manager().get_default_pipeline_config()
        )
        feedback_interactive_st = gr.State(False)
        output_dirty_st = gr.State(True)
        feedback_likert_content_st = gr.State(dict())
        toxicity_example_data = _load_toxicity_example_data()
        toxicity_example_data_st = gr.State(toxicity_example_data)
        user_input_source_st = gr.State("")

        with gr.Tabs(
            selected="detector_tab" if tw_approved_value else "tw_tab"
        ) as tabs:
            # Create tabs
            detector_tab = create_detection_tab(
                tw_approved=tw_approved,
                tw_approved_value=tw_approved_value,
                pipeline_config_state=pipeline_config_state,
                result_state=result_state,
                feedback_interactive_st=feedback_interactive_st,
                output_dirty_st=output_dirty_st,
                feedback_likert_content_st=feedback_likert_content_st,
                toxicity_example_data_st=toxicity_example_data_st,
                user_input_source_st=user_input_source_st,
            )

            config_tab = create_config_tab(
                tw_approved=tw_approved,
                tw_approved_value=tw_approved_value,
                pipeline_config_state=pipeline_config_state,
            )

            tw_tab = create_agreement_tab(
                tw_approved=tw_approved,
                detector_tab=detector_tab,
                config_tab=config_tab,
                tabs=tabs,
            )

    return demo


def launch_app(
    config_path=None,
    config_type="app",
    server_name="127.0.0.1",
    server_port=7860,
    share=False,
):
    """Launch the Gradio app with specified configuration.

    Args:
        config_path: Path to configuration file. If None, uses default.
        config_type: Either "app" or "pipeline". Determines how to load config.
        server_name: Server name/host (default: 127.0.0.1)
        server_port: Server port (default: 7860)
        share: Whether to create a public shareable link (default: False)
    """
    # Initialize app config
    logger.info(f"Loading {config_type} config from: {config_path}")
    init_app_config(config_path, config_type)

    # Create the Gradio demo
    demo = create_demo()

    # Launch the server
    logger.info(f"Starting Gradio app on {server_name}:{server_port}")
    demo.launch(
        show_error=True,
        server_name=server_name,
        server_port=server_port,
        share=share,
    )



"""Configuration loader for the Gradio app."""

import os
import gradio as gr
from pathlib import Path

from toxicity_detector import ConfigManager
from toxicity_detector.config import AppConfig, PipelineConfig

# Global variables to hold app config (will be initialized by init_app_config)
app_config: AppConfig | None = None
config_manager: ConfigManager | None = None


def _config_manager() -> ConfigManager:
    """Get the initialized ConfigManager instance."""
    if config_manager is None:
        raise RuntimeError(
            "ConfigManager not initialized. Call init_app_config() first."
        )
    return config_manager


def _app_config() -> AppConfig:
    """Get the initialized AppConfig instance."""
    if app_config is None:
        raise RuntimeError("AppConfig not initialized. Call init_app_config() first.")
    return app_config


def init_app_config(config_path=None, config_type="app"):
    """Initialize app configuration and config manager.

    Args:
        config_path: Path to configuration file. If None, uses default path.
        config_type: Either "app" for AppConfig or "pipeline" for PipelineConfig.
                     If "pipeline", creates AppConfig with defaults.
    """
    global app_config, config_manager

    # Load app config file path from environment variable if not provided
    if config_path is None:
        config_path = os.getenv("TOXICITY_DETECTOR_APP_CONFIG_FILE")
        if config_path is None:
            raise ValueError(
                "No config_path provided and "
                "TOXICITY_DETECTOR_APP_CONFIG_FILE not set as env!"
            )
        if config_type == "pipeline":
            raise ValueError(
                "config_type 'pipeline' cannot be used "
                "when config_path is from env variable!"
            )

    if config_type == "app":
        app_config = AppConfig.from_file(config_path)
        # re-init None values from pipeline config values and validate
        pipeline_config = PipelineConfig.from_file(
            os.path.join(
                app_config.config_path,
                app_config.pipeline_config_file
            )
        )
        if app_config.local_serialization is None:
            if pipeline_config.local_serialization is None:
                raise gr.Error(
                    "local_serialization must be set in either "
                    "app config or pipeline config!"
                )
            app_config.local_serialization = pipeline_config.local_serialization
        if app_config.local_base_path is None:
            if (
                app_config.local_serialization
                and pipeline_config.local_base_path is None
            ):
                raise gr.Error(
                    "local_base_path must be set in either "
                    "app config or pipeline config when "
                    "local_serialization is True!"
                )
            app_config.local_base_path = pipeline_config.local_base_path
        if app_config.hf_base_path is None:
            if (
                not app_config.local_serialization
                and pipeline_config.hf_base_path is None
            ):
                raise gr.Error(
                    "hf_base_path must be set in either "
                    "app config or pipeline config when "
                    "local_serialization is False!"
                )
            app_config.hf_base_path = pipeline_config.hf_base_path
        if app_config.hf_key_name is None:
            app_config.hf_key_name = pipeline_config.hf_key_name
        if app_config.env_file is not None:
            app_config.env_file = pipeline_config.env_file
            app_config.load_env_file()

        config_manager = ConfigManager(
            app_config.local_serialization,
            app_config.local_base_path,
            app_config.hf_base_path,
            app_config.hf_key_name,
            app_config.config_path,
            app_config.pipeline_config_file,
            app_config.pipeline_config_version,
        )
    elif config_type == "pipeline":
        # Load pipeline config, create AppConfig with defaults
        # and initialize ConfigManager with values from pipeline config
        pipeline_config = PipelineConfig.from_file(config_path)
        # Create minimal AppConfig with defaults
        # (We do not need pass serialization configuration since we
        # pass it directly to the ConfigManager.)
        app_config = AppConfig(pipeline_config_file=Path(config_path).name)
        # Initialize ConfigManager
        config_manager = ConfigManager(
            pipeline_config.local_serialization,
            pipeline_config.local_base_path,
            pipeline_config.hf_base_path,
            pipeline_config.hf_key_name,
            app_config.config_path,  # defaults to "configs"
            Path(config_path).name,  # setting the passed pipeline config as default
            pipeline_config.config_version,  # version from loaded pipeline config
        )
    else:
        raise ValueError(
            f"Invalid config_type: {config_type}. Must be 'app' or 'pipeline'."
        )
